@@ -10,9 +10,11 @@
 #ifdef RKTIO_SYSTEM_UNIX
 # include <unistd.h>
 # include <netinet/in.h>
+# include <netinet/tcp.h>
 # include <netdb.h>
 # include <sys/socket.h>
 # include <sys/types.h>
+# include <sys/time.h>
 # include <fcntl.h>
 # include <errno.h>
 # define TCP_SOCKSENDBUF_SIZE 32768
@@ -209,9 +211,7 @@ struct rktio_addrinfo_t {
 # define rktio_AI_PASSIVE AI_PASSIVE 
 # define do_getaddrinfo(n, s, h, res) getaddrinfo(n, s, RKTIO_AS_ADDRINFO(h), RKTIO_AS_ADDRINFO_PTR(res))
 # define do_freeaddrinfo freeaddrinfo
-# ifdef RKTIO_SYSTEM_WINDOWS
-#  define do_gai_strerror gai_strerrorA
-# else
+# ifndef RKTIO_SYSTEM_WINDOWS
 #  define do_gai_strerror gai_strerror
 # endif
 #else
@@ -291,7 +291,7 @@ struct rktio_addrinfo_lookup_t {
 # ifdef RKTIO_SYSTEM_WINDOWS
   HANDLE done_sema;
 # else
-  int done_fd[2];
+  intptr_t done_fd[2];
 # endif
 
   /* For chaining active requests: */
@@ -517,8 +517,10 @@ static rktio_addrinfo_lookup_t *start_lookup(rktio_t *rktio, rktio_addrinfo_look
   
   if (!rktio->ghbn_started) {
     rktio->ghbn_run = 1;
-    if (!ghbn_init(rktio))
+    if (!ghbn_init(rktio)) {
+      free_lookup(lookup);
       return NULL;
+    }
     rktio->ghbn_started = 1;
   }
 
@@ -532,7 +534,7 @@ static rktio_addrinfo_lookup_t *start_lookup(rktio_t *rktio, rktio_addrinfo_look
     }
   }
 # else
-  if (pipe(lookup->done_fd)) {
+  if (rktio_make_os_pipe(rktio, lookup->done_fd, RKTIO_NO_INHERIT_INPUT | RKTIO_NO_INHERIT_OUTPUT)) {
     get_posix_error();
     free_lookup(lookup);
     return NULL;
@@ -722,6 +724,15 @@ rktio_addrinfo_lookup_t *rktio_start_addrinfo_lookup(rktio_t *rktio,
   if (passive) {
     RKTIO_AS_ADDRINFO(hints)->ai_flags |= rktio_AI_PASSIVE;
   }
+  /* On modern systems, `AI_ADDRCONFIG` and `AI_V4MAPPED` tend to be
+     the defaults when `hints` is NULL, so add them if they seem
+     available. */
+#ifdef AI_ADDRCONFIG
+  RKTIO_AS_ADDRINFO(hints)->ai_flags |= AI_ADDRCONFIG;
+#endif
+#ifdef AI_V4MAPPED
+  RKTIO_AS_ADDRINFO(hints)->ai_flags |= AI_V4MAPPED;
+#endif
   if (tcp) {
     RKTIO_AS_ADDRINFO(hints)->ai_socktype = SOCK_STREAM;
 # ifndef PROTOENT_IS_INT
@@ -748,9 +759,18 @@ void rktio_addrinfo_free(rktio_t *rktio, rktio_addrinfo_t *a)
   do_freeaddrinfo(RKTIO_AS_ADDRINFO(a));
 }
 
-const char *rktio_gai_strerror(int errnum)
+const char *rktio_gai_strerror(rktio_t *rktio, int errnum)
 {
+#ifdef RKTIO_SYSTEM_WINDOWS
+  char *s;
+  s = NARROW_PATH_copy(gai_strerrorW(errnum));
+  if (rktio->last_err_str)
+    free(rktio->last_err_str);
+  rktio->last_err_str = s;
+  return s;
+#else
   return do_gai_strerror(errnum);
+#endif
 }
 
 /*========================================================================*/
@@ -959,6 +979,18 @@ int rktio_socket_shutdown(rktio_t *rktio, rktio_fd_t *rfd, int mode)
   return 1;
 }
 
+int rktio_tcp_nodelay(rktio_t *rktio, rktio_fd_t *rfd, rktio_bool_t enable)
+{
+  rktio_socket_t s = rktio_fd_socket(rktio, rfd);
+  int nd = (enable ? 1 : 0), r;
+  r = setsockopt(s, IPPROTO_TCP, TCP_NODELAY, &nd, sizeof(nd));
+  if (r) {
+    get_socket_error();
+    return 0;
+  }
+  return 1;
+}
+
 int rktio_socket_poll_write_ready(rktio_t *rktio, rktio_fd_t *rfd)
 {
 #ifdef RKTIO_SYSTEM_UNIX
@@ -1055,7 +1087,9 @@ intptr_t rktio_socket_read(rktio_t *rktio, rktio_fd_t *rfd, char *buffer, intptr
 {
   rktio_socket_t s = rktio_fd_socket(rktio, rfd);
   int rn;
-  
+
+  len = LIMIT_REQUEST_SIZE(len);
+
   do {
     rn = recv(s, buffer, len, 0);
   } while ((rn == -1) && NOT_WINSOCK(errno == EINTR));
@@ -1095,7 +1129,7 @@ static intptr_t do_socket_write(rktio_t *rktio, rktio_fd_t *rfd, const char *buf
 
   while (1) {
     if (addr) {
-      /* Use first address that dosn't result in a bad-address error: */
+      /* Use first address that doesn't result in a bad-address error: */
       for (; addr; addr = (rktio_addrinfo_t *)RKTIO_AS_ADDRINFO(addr)->ai_next) {
         do {
           sent = sendto(s, buffer, len, 0,
@@ -1133,7 +1167,7 @@ static intptr_t do_socket_write(rktio_t *rktio, rktio_fd_t *rfd, const char *buf
 
 intptr_t rktio_socket_write(rktio_t *rktio, rktio_fd_t *rfd, const char *buffer, intptr_t len)
 {
-  return do_socket_write(rktio, rfd, buffer, len, NULL);
+  return do_socket_write(rktio, rfd, buffer, LIMIT_REQUEST_SIZE(len), NULL);
 }
 
 /*========================================================================*/
@@ -1170,11 +1204,21 @@ static rktio_connect_t *try_connect(rktio_t *rktio, rktio_connect_t *conn)
 {
   struct rktio_addrinfo_t *addr;
   rktio_socket_t s;
-  
+
+  rktio_cloexec_lock();
+
   addr = conn->addr;
   s = socket(RKTIO_AS_ADDRINFO(addr)->ai_family,
              RKTIO_AS_ADDRINFO(addr)->ai_socktype,
              RKTIO_AS_ADDRINFO(addr)->ai_protocol);
+
+  if (s != INVALID_SOCKET)
+    rktio_fd_cloexec(s);
+  else
+    get_socket_error();
+
+  rktio_cloexec_unlock();
+
   if (s != INVALID_SOCKET) {
     int status, inprogress;
     if (!conn->src
@@ -1215,7 +1259,6 @@ static rktio_connect_t *try_connect(rktio_t *rktio, rktio_connect_t *conn)
     }
   }
   
-  get_socket_error();
   return NULL;
 }
 
@@ -1365,12 +1408,19 @@ rktio_listener_t *rktio_listen(rktio_t *rktio, rktio_addrinfo_t *src, int backlo
 	}
 #endif
 
+        rktio_cloexec_lock();
+
 	s = socket(RKTIO_AS_ADDRINFO(addr)->ai_family,
                    RKTIO_AS_ADDRINFO(addr)->ai_socktype,
                    RKTIO_AS_ADDRINFO(addr)->ai_protocol);
+
         if (s != INVALID_SOCKET)
+          rktio_fd_cloexec(s);
+        else
           get_socket_error();
 
+        rktio_cloexec_unlock();
+
 #ifdef RKTIO_TCP_LISTEN_IPV6_ONLY_SOCKOPT
 	if (s == INVALID_SOCKET) {
 	  /* Maybe it failed because IPv6 is not available: */
@@ -1618,8 +1668,7 @@ void rktio_poll_add_accept(rktio_t *rktio, rktio_listener_t *listener, rktio_pol
 
 rktio_fd_t *rktio_accept(rktio_t *rktio, rktio_listener_t *listener)
 {
-  int
-    ready_pos;
+  int ready_pos;
   rktio_socket_t s, ls;
   rktio_sockopt_len_t l;
   char tcp_accept_addr[RKTIO_SOCK_NAME_MAX_LEN];
@@ -1634,11 +1683,20 @@ rktio_fd_t *rktio_accept(rktio_t *rktio, rktio_listener_t *listener)
 
   l = sizeof(tcp_accept_addr);
 
+  rktio_cloexec_lock();
+
   do {
     s = accept(ls, (struct sockaddr *)tcp_accept_addr, &l);
   } while ((s == -1) && NOT_WINSOCK(errno == EINTR));
 
-  if (s != INVALID_SOCKET) {    
+  if (s != INVALID_SOCKET)
+    rktio_fd_cloexec(s);
+  else
+    get_socket_error();
+
+  rktio_cloexec_unlock();
+
+  if (s != INVALID_SOCKET) {
 # ifdef RKTIO_SYSTEM_UNIX
     RKTIO_WHEN_SET_SOCKBUF_SIZE(int size = TCP_SOCKSENDBUF_SIZE);
     RKTIO_WHEN_SET_SOCKBUF_SIZE(setsockopt(s, SOL_SOCKET, SO_SNDBUF, (char *)&size, sizeof(int)));
@@ -1647,7 +1705,6 @@ rktio_fd_t *rktio_accept(rktio_t *rktio, rktio_listener_t *listener)
     return rktio_system_fd(rktio, s, (RKTIO_OPEN_SOCKET | RKTIO_OPEN_INIT | RKTIO_OPEN_OWN
 				      | RKTIO_OPEN_READ | RKTIO_OPEN_WRITE)); 
   } else {
-    get_socket_error();
     return NULL;
   }
 }
@@ -1736,6 +1793,8 @@ char **rktio_listener_address(rktio_t *rktio, rktio_listener_t *lnr)
 rktio_fd_t *rktio_udp_open(rktio_t *rktio, rktio_addrinfo_t *addr, int family)
 {
   rktio_socket_t s;
+
+  rktio_cloexec_lock();
   
   if (addr)
     s = socket(RKTIO_AS_ADDRINFO(addr)->ai_family,
@@ -1744,10 +1803,15 @@ rktio_fd_t *rktio_udp_open(rktio_t *rktio, rktio_addrinfo_t *addr, int family)
   else
     s = socket(family, SOCK_DGRAM, 0);
 
-  if (s == INVALID_SOCKET) {
+  if (s != INVALID_SOCKET)
+    rktio_fd_cloexec(s);
+  else
     get_socket_error();
+
+  rktio_cloexec_unlock();
+
+  if (s == INVALID_SOCKET)
     return NULL;
-  }
 
   return rktio_system_fd(rktio, s, RKTIO_OPEN_SOCKET | RKTIO_OPEN_UDP | RKTIO_OPEN_INIT);
 }
@@ -1915,7 +1979,7 @@ int rktio_udp_set_receive_buffer_size(rktio_t *rktio, rktio_fd_t *rfd, int size)
 int rktio_udp_get_multicast_loopback(rktio_t *rktio, rktio_fd_t *rfd)
 {
   rktio_socket_t s = rktio_fd_socket(rktio, rfd);
-  u_char loop;
+  unsigned char loop;
   rktio_sockopt_len_t loop_len = sizeof(loop);
   int status;
   
@@ -1931,7 +1995,7 @@ int rktio_udp_get_multicast_loopback(rktio_t *rktio, rktio_fd_t *rfd)
 int rktio_udp_set_multicast_loopback(rktio_t *rktio, rktio_fd_t *rfd, int on)
 {
   rktio_socket_t s = rktio_fd_socket(rktio, rfd);
-  u_char loop = (on ? 1 : 0);
+  unsigned char loop = (on ? 1 : 0);
   rktio_sockopt_len_t loop_len = sizeof(loop);
   int status;
   
@@ -1944,10 +2008,42 @@ int rktio_udp_set_multicast_loopback(rktio_t *rktio, rktio_fd_t *rfd, int on)
     return 1;
 }
 
+int rktio_udp_get_ttl(rktio_t *rktio, rktio_fd_t *rfd)
+{
+  rktio_socket_t s = rktio_fd_socket(rktio, rfd);
+  int ttl;
+  rktio_sockopt_len_t ttl_len = sizeof(ttl);
+  int status;
+
+  status = getsockopt(s, IPPROTO_IP, IP_TTL, (void *)&ttl, &ttl_len);
+
+  if (status) {
+    get_socket_error();
+    return RKTIO_PROP_ERROR;
+  } else
+    return ttl;
+}
+
+int rktio_udp_set_ttl(rktio_t *rktio, rktio_fd_t *rfd, int ttl_val)
+{
+  rktio_socket_t s = rktio_fd_socket(rktio, rfd);
+  int ttl = ttl_val;
+  rktio_sockopt_len_t ttl_len = sizeof(ttl);
+  int status;
+
+  status = setsockopt(s, IPPROTO_IP, IP_TTL, (void *)&ttl, ttl_len);
+
+  if (status) {
+    get_socket_error();
+    return 0;
+  } else
+    return 1;
+}
+
 int rktio_udp_get_multicast_ttl(rktio_t *rktio, rktio_fd_t *rfd)
 {
   rktio_socket_t s = rktio_fd_socket(rktio, rfd);
-  u_char ttl;
+  unsigned char ttl;
   rktio_sockopt_len_t ttl_len = sizeof(ttl);
   int status;
   
@@ -1963,7 +2059,7 @@ int rktio_udp_get_multicast_ttl(rktio_t *rktio, rktio_fd_t *rfd)
 int rktio_udp_set_multicast_ttl(rktio_t *rktio, rktio_fd_t *rfd, int ttl_val)
 {
   rktio_socket_t s = rktio_fd_socket(rktio, rfd);
-  u_char ttl = ttl_val;
+  unsigned char ttl = ttl_val;
   rktio_sockopt_len_t ttl_len = sizeof(ttl);
   int status;
   
